/**
 * SessionDetailFragment.java: Practice session detail, and recording.
 *            Copyright (C) 2012 donfede@casagrau.org
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, write to the Free Software Foundation, Inc., 59 Temple
 * Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * created 2012-09-23 donfede@casagrau.org
 *
 * $Id: SessionDetailFragment.java,v 1.35 2012-11-06 01:29:10 donfede Exp $
 */

package org.casagrau.mpq.mpq_android;

import java.io.File;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;

import org.casagrau.mpq.mpq_android.SessionsDBProvider.SessionsDB;

import android.app.Activity;
import android.content.ContentUris;
import android.content.ContentValues;
import android.content.SharedPreferences;
import android.database.Cursor;
import android.media.MediaPlayer;
import android.media.MediaRecorder;
import android.net.Uri;
import android.os.Bundle;
import android.os.Environment;
import android.preference.PreferenceManager;
import android.support.v4.app.DialogFragment;
import android.support.v4.app.Fragment;
import android.support.v4.app.LoaderManager;
import android.support.v4.content.CursorLoader;
import android.support.v4.content.Loader;
import android.support.v4.widget.SimpleCursorAdapter;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.View.OnClickListener;
import android.view.ViewGroup;
import android.widget.Button;
import android.widget.EditText;
import android.widget.Spinner;
import android.widget.TextView;
import android.widget.Toast;

/**
 * Fragment with GUI for Session Detail. This screen will be active while a user
 * is in a practice session, and allow the user to review and edit details about
 * a practice session.
 */
public class SessionDetailFragment extends Fragment implements
		LoaderManager.LoaderCallbacks<Cursor> , 
		MPQGenDialogFragment.MPQGenDialogListener {
	private static final String LOGTAG = "SessionDetailFragment";
	private static final String BNDL_SESSION_ID = "session_id"; // Bundle handle to SesId
	private static final String BNDL_SESSION_STATE = "session_state"; // Bundle handle to SesState
	
	static final String ARG_DETAIL_FLAG = "detail_session_flag"; // Argument handle
	static final long ARG_DETAIL_FLAG_ADD = -1;					 // Negative argument value code to add a session
	static final long ARG_DETAIL_FLAG_CURRENT = 0;				 // Zero value argument value code to show the current or most recent session 

	private static final int DLG_SAVE_DISCARD = 10; 	// Identifier code for the save discard dialog.

	
	static final int FLG_SES_STATE_VIEW = 0;					 // Session state is in View mode.
	static final int FLG_SES_STATE_EDIT = 1;					 // Session state is in Edit mode.
// FIXME: review if a create/recording mode state is necessary... - tidy up mSesState appropriately.	
	
	// Unique integer identifiers for Loaders; rather warty... :(
	private static final int LDR_GOALS = 0;
	private static final int LDR_INSTRUMENTS = 1;
	private static final int LDR_SESSIONS = 2;

	// Adapters to bind to CursorLoaders
	SimpleCursorAdapter mAdapterGoals;
	SimpleCursorAdapter mAdapterInstruments;
	SimpleCursorAdapter mAdapterSessions;
	
	// Handles to various UI widgets, to save re-querying them.  FIXME: these should probably go into an array
	private static long mSesId = ARG_DETAIL_FLAG_CURRENT; // Identifier of the current practice session for which detail is being displayed and processed. // FIXME: keeping this static, to maintain user consistency when returning to the tab
//	private long mSesId = ARG_DETAIL_FLAG_CURRENT; // Identifier of the current practice session for which detail is being displayed and processed. // FIXME: keeping this static, to maintain user consistency when returning to the tab
	private long mSesCount;	 // Number of rows of practice sessions.
	private long mSesMin;	 // Minimum row id value for practice sessions.
	private long mSesMax;    // Max row id value for practice sessions.
	private int mSesState;	 // Flag to track session state (View, Edit, "Create"[not sure if latter is needed)
	/**
	 * Class to keep track of meta data associated with a UI widget element. 
	 */
	private class SessionDetailWidgetMeta {
		// Widget Type constants
		private static final int WT_INVALID = -1;
		private static final int WT_SPINNER = 0;
		private static final int WT_EDITTEXT = 1;
		private static final int WT_TEXTVIEW = 2;

		private String mDBCol; // for now - access the members directly from the calling object... try to restrict access to within this class using priavte (might be off here)
		private int mDBColPos;
		private int mDBColType; // corresponds with CURSOR.FIELD_TYPE
		private int mResourceID;
		private View mUIWidget;
		private int mWidgetType; // maybe use Cursor.FIELD_TYPE_xxx constants ... but still need widget type... 
//FIXME: should this object include : A)cursor method to extract? 
		/** 
		 * Constructor that populates empty values
		 */
		SessionDetailWidgetMeta() {
			mDBCol = "";
			mDBColPos = -1;
			mResourceID = -1;
			mUIWidget = null;
			mWidgetType = WT_INVALID;
		}
	} // END class SessionDetailWidgets
	// An array to keep track of widget-DB data. 
	private List<SessionDetailWidgetMeta> mDetailWidgets = new ArrayList<SessionDetailWidgetMeta>(); // FIXME: not sure if we can instantiate this object here...
	
	
	// recording member variables (FIXME: these should be relocated into a
	// recording/transport_control class)
	private String mFileName = null;
	private MediaRecorder mRecorder = null;
	private MediaPlayer mPlayer = null;
	private boolean mStartRecording = true;
	private boolean mStartPlaying = true;
	private Button mRecordButton = null;
	private Button mPlayButton = null;
	
	private Button mPrevButton = null;
	private Button mNextButton = null;
	private Button mEditLockButton = null;

	// Parameterless constructor, required by framework
	public SessionDetailFragment() {
		super();
	}

	@Override
	public void onAttach(Activity a) {
		super.onAttach(a);
		Log.d(LOGTAG, "onAttach() 00");
		Log.d(LOGTAG, "onAttach() 10 activity:" + a.getLocalClassName());
	}
	
	@Override
	public void onCreate(Bundle savedInstanceState) {
		super.onCreate(savedInstanceState);
		Log.d(LOGTAG, "onCreate() 00");		

		if (savedInstanceState != null) {
			Log.d(LOGTAG, "onCreate() 10   found bundle:" + savedInstanceState.size());		
		} else Log.d(LOGTAG, "onCreate() 11   NO bundle");
	}
	
	
	/**
	 * Create the view for this fragment - loading XML layout; parsing fragment arguments.
	 */
	@Override
	public View onCreateView(LayoutInflater inflater, ViewGroup container,
			Bundle savedInstanceState) {
//FIXME: save/retrieve detail mSesId in bundle 		
		Log.d(LOGTAG, "onCreateView() 00 mSesId:" + mSesId);		

		long lflag = SessionDetailFragment.ARG_DETAIL_FLAG_CURRENT;
		if (getArguments() != null) {
			lflag = getArguments().getLong(ARG_DETAIL_FLAG, 
					SessionDetailFragment.ARG_DETAIL_FLAG_CURRENT); 
			Log.d(LOGTAG, "onCreateView() 10   found args lflag:" + lflag);		
		} else Log.d(LOGTAG, "onCreateView() 11   NO args   default lflag:" + lflag);
		
		// Get the previous state, or initialize state to viewing.
		if (savedInstanceState != null) {
			mSesState = savedInstanceState.getInt(BNDL_SESSION_STATE);
			Log.d(LOGTAG, "onCreateView() 21   found BNDL State:" + mSesState);		
		} else {
			mSesState = FLG_SES_STATE_VIEW;
			Log.d(LOGTAG, "onCreateView() 24   init NO BNDL State:" + mSesState);
		}
		
		if (lflag < 0 ) {
			// open a new fragment
			mSesId = newSessionDetailData();
			mSesState = FLG_SES_STATE_EDIT;
			Log.d(LOGTAG, "onCreateView() 30   add new mSesId:" + mSesId);
		} else if (lflag > 0) {
			// (try to) open specific session
			mSesId = lflag;
			Log.d(LOGTAG, "onCreateView() 60   get try mSesId:" + mSesId + " state:" + mSesState);
		} else if ( (lflag == 0) && (savedInstanceState != null) ) { 
			// open current or most recent fragment
        	mSesId = savedInstanceState.getLong(BNDL_SESSION_ID);
			Log.d(LOGTAG, "onCreateView() 30   savedInstance not null mSesId:" + mSesId + " state:" + mSesState);		
        } 
		
		// sanity check and get next or get prev get max
		int i = 0;
		if (mSesId > 0)
			i = getSessionDetailIdCount(mSesId);
		if ( (i != 1) || (mSesId == 0) ) {
			updateSessionDetailCount();
			mSesId = mSesMax;
	    }
		
		
		Log.d(LOGTAG, "onCreateView() 99 mSesId:" + mSesId);
		return inflater.inflate(R.layout.fragment_sessiondetail, container, false);
	}
	
	/**
	 * as the Activity is created, prepare the LoaderManager and associate screen widgets with Adapters.
	 */
	@SuppressWarnings("deprecation") // benign complaints about a temporary null cursor used...
	@Override
	public void onActivityCreated(Bundle savedInstanceState) {
		Log.d(LOGTAG, "onActivityCreated() 00   ");
		super.onActivityCreated(savedInstanceState);
		// Get the previous state, or initialize state to viewing.
		if (savedInstanceState != null) {
			Log.d(LOGTAG, "onActivityCreated() 11   found BNDL State:" + savedInstanceState.size());		
		} else {
			Log.d(LOGTAG, "onActivityCreated() 14   init NO BNDL ");
		}
		Activity lActivity = getActivity();
		
		// Get handles other widgets and store in member attributes.
		// Ideally these would only be defined here - but the current extraction is a bit clumsy FIXME: polish/redo
		int lcol = 0;
		// RowId
		SessionDetailWidgetMeta wm = new SessionDetailWidgetMeta();
		wm.mDBColPos = lcol++;
		wm.mDBCol = SessionsDB._ID;
		mDetailWidgets.add(wm);
		// DateStamp
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_datetime;
		wm.mUIWidget = (EditText) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_DATE_STAMP;
		mDetailWidgets.add(wm);
		// length
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_length;
		wm.mUIWidget = (EditText) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_LENGTH_TOTAL;
		mDetailWidgets.add(wm);
		// noise length
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_noiselength;
		wm.mUIWidget = (EditText) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_LENGTH_NOISE;
		mDetailWidgets.add(wm);
		// audio filename
		wm = new SessionDetailWidgetMeta();
// error-taking over screen, so using a separate TextView label instead.		wm.mResourceID = R.id.wave_form;
		wm.mResourceID = R.id.label_audio_file;
		wm.mUIWidget = (TextView) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_TEXTVIEW;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_AUDIO_FILE;
		mDetailWidgets.add(wm);		
		// goal - do later - pull down from above... populate spinner - FIXME: possible race condition, with spinner contents being populated...
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_spinner_goal;
		wm.mUIWidget = (Spinner) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_SPINNER;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_GOAL_ID;
		mDetailWidgets.add(wm);
		// notes
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_text_notes;
		wm.mUIWidget = (EditText) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_EDITTEXT;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_NOTES;
		mDetailWidgets.add(wm);
		// instrument - do later - see goal spinner notes above
//FIXME: bring down from above...		
		wm = new SessionDetailWidgetMeta();
		wm.mResourceID = R.id.detail_spinner_instrument;
		wm.mUIWidget = (Spinner) lActivity.findViewById(wm.mResourceID);
		wm.mWidgetType = SessionDetailWidgetMeta.WT_SPINNER;
		wm.mDBColPos = lcol++; 
		wm.mDBCol = SessionsDB.PRACTICE_INSTRUMENT_ID;
		mDetailWidgets.add(wm);
		
		// Initialize adapters to "empty", they will later be updated by the Loader.
		mAdapterGoals = new SimpleCursorAdapter(getActivity(),
				android.R.layout.simple_spinner_item, null,
				new String[] { SessionsDB.GOAL_NAME },
				new int[] { android.R.id.text1 });
		mAdapterGoals.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);
		mAdapterInstruments = new SimpleCursorAdapter(getActivity(),
				android.R.layout.simple_spinner_item, null,
				new String[] { SessionsDB.INSTRUMENT_NAME },
				new int[] { android.R.id.text1 });
		mAdapterInstruments.setDropDownViewResource(android.R.layout.simple_spinner_dropdown_item);

		

		// Prepare the Loaders, triggering callback methods. - this should occur after the above widget meta data is populated
		getLoaderManager().initLoader(LDR_INSTRUMENTS, null, this);
		getLoaderManager().initLoader(LDR_GOALS, null, this);
		// connect Spinner controls with CursorAdapter
		Spinner spinnerInstruments = (Spinner) lActivity.findViewById(R.id.detail_spinner_instrument);
		spinnerInstruments.setAdapter(mAdapterInstruments);
		Spinner spinnerGoals = (Spinner) lActivity.findViewById(R.id.detail_spinner_goal);
		spinnerGoals.setAdapter(mAdapterGoals);
		
		getLoaderManager().initLoader(LDR_SESSIONS, null, this);			

		Log.d(LOGTAG, "onActivityCreated() 99");
	} // END onActivityCreated()

	@Override
	public void onResume() {
		super.onResume();
		Activity lact = getActivity(); // local variable with activity context
		Log.d(LOGTAG, "onResume() 00");

		// bind listeners to transport control buttons
		// bind rec listener
		mRecordButton = (Button) lact.findViewById(R.id.buttonDetailTCRec);
		mRecordButton.setOnClickListener(clickerRec);
		// bind play listener
		mPlayButton = (Button) lact.findViewById(R.id.buttonDetailTC);
		mPlayButton.setOnClickListener(clickerPlay);
		// bind previous and next scrolling buttons - these are likely temporary 
		mPrevButton = (Button) lact.findViewById(R.id.buttonDetailScrollPrev);
		mPrevButton.setOnClickListener(clickerPrev); 
		mNextButton = (Button) lact.findViewById(R.id.buttonDetailScrollNext);
		mNextButton.setOnClickListener(clickerNext);
		mEditLockButton = (Button) lact.findViewById(R.id.buttonDetailEditLock);
		mEditLockButton.setOnClickListener(clickerEditLock);
		// Update EditLock button text, and other UI widgets enabled state. // FIXME: this is feeling like a function (button state toggles)
		if (FLG_SES_STATE_VIEW == mSesState) {
			mEditLockButton.setText(getString(R.string.button_detail_unlock));
			mPrevButton.setEnabled(true);
			mNextButton.setEnabled(true);
		} else {
			mEditLockButton.setText(getString(R.string.button_detail_save)); // FIXME: probably need a better string identifier name
			mPrevButton.setEnabled(false);
			mNextButton.setEnabled(false);
		}
		updateSessionDetailWidgets();
		Log.d(LOGTAG, "onResume() 99");
	}

	@Override
	public void onPause() {
		super.onPause();
		Log.d(LOGTAG, "onPause() 00");
		// Cleanup record/tc environment. FIXME: this belongs in a class
		if (mRecorder != null) {
			mRecorder.release();
			mRecorder = null;
		}
		if (mPlayer != null) {
			mPlayer.release();
			mPlayer = null;
		}
		mRecordButton = null;
		mPlayButton = null;
		Log.d(LOGTAG, "onPause() 99");
	}
	
	@Override
	public void onDetach() {
		super.onDetach();
		Log.d(LOGTAG, "onDetach() 00");
	}
	
    /**
     * Save state of the app - mSesId only for now.
     */
    @Override
    public void onSaveInstanceState(Bundle outState) {
     	outState.putLong(BNDL_SESSION_ID , mSesId);
    	outState.putInt(BNDL_SESSION_STATE, mSesState);
    	Log.d(LOGTAG, "DEBUG onSaveInstanceState() 00");
    	Log.d(LOGTAG, "DEBUG onSaveInstanceState() 10 outState.size:" + outState.size());
    	// Always call the superclass so it can save the view hierarchy state
    	super.onSaveInstanceState(outState);
    }	

	// play listener // FIXME: no ability to pause ... only stop...
	OnClickListener clickerPlay = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "Play onClick() 00");
			onPlay(mStartPlaying);
			if (mStartPlaying) {
				mPlayButton.setText(getString(R.string.button_detail_tcstop)); // FIXME: concern about scope -- will this callback be able to access class members???
				Log.d(LOGTAG, "Play onClick() 11");
			} else {
				mPlayButton.setText(getString(R.string.button_detail_tcplay));
				Log.d(LOGTAG, "Play onClick() 15");
			}
			mStartPlaying = !mStartPlaying;
			Log.d(LOGTAG, "Play onClick() 99");
		}

	}; // END OnClickListener clickerPlay
	// record listener
	OnClickListener clickerRec = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "Rec onClick() 00");
			onRecord(mStartRecording);
			if (mStartRecording) {
				mRecordButton.setText(getString(R.string.button_detail_tcstop)); // FIXME: concern about scope -- will this callback be able to access class members???
			} else {
				mRecordButton.setText(getString(R.string.button_detail_tcrec));
			}
			mStartRecording = !mStartRecording;
			Log.d(LOGTAG, "Rec onClick() 99");
		}
	}; // END OnClickListener clickerRec
	
	// previous button listener
	OnClickListener clickerPrev = new OnClickListener() {
		public void onClick(View v) {	
			Log.d(LOGTAG, "Prev onClick() 00");

			// SELECT * FROM dictionary WHERE _id = (select max(_id) from dictionary where _id < ?)
//			lNewUri = getActivity().getContentResolver().insert(SessionsDB.CONTENT_PRACTICES_URI, val);
			Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_PREV_URI, mSesId);
			Log.d(LOGTAG, "Prev onClick() 10 luri:" + luri);
			Cursor lcursor = getActivity().getContentResolver().query(luri, null, null, null, null);
			long lrow = lcursor.getLong(0);
			lcursor.close();
			Log.d(LOGTAG, "Prev onClick() 20 lrow:" + lrow + " *** ");

			if (lrow != 0) { // Check if lrow is valid.
				mSesId = lrow;
				if (mSesId < mSesMin) {
					mSesId = mSesMin;
					Toast.makeText(getActivity(),getString(R.string.button_detail_prev_msg), Toast.LENGTH_LONG).show(); // FIXME: this is likely superfluous
				}
				updateSessionDetailData();
			} else {
				Toast.makeText(getActivity(),getString(R.string.button_detail_prev_msg), Toast.LENGTH_LONG).show();
			}
		}
	}; // END OnClickListener clickerPrev
	// next button listener
	OnClickListener clickerNext = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "Next onClick() 00");
			Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_NEXT_URI, mSesId);
			Log.d(LOGTAG, "Next onClick() 10 luri:" + luri);
			Cursor lcursor = getActivity().getContentResolver().query(luri, null, null, null, null);
			long lrow = lcursor.getLong(0);
			lcursor.close();
			Log.d(LOGTAG, "Next onClick() 20 lrow:" + lrow + " *** ");
			
			if (lrow > mSesId) { // Only do anything if we have a value worth processing.
				mSesId = lrow;  			
				if (mSesId > mSesMax) { // But confirm the value is within bounds.
					mSesId = mSesMax;
					Toast.makeText(getActivity(),getString(R.string.button_detail_next_msg), Toast.LENGTH_LONG).show(); // FIXME: this is likely superfluous
				}
				updateSessionDetailData();
			} else {
				Toast.makeText(getActivity(),getString(R.string.button_detail_next_msg), Toast.LENGTH_LONG).show();				
			}
		}
	}; // END OnClickListener clickerNext
	
	/**
	 * Respond to edit/unlock button event.
	 *   Change state for the current session detail.
	 *     Set a flag to save data in a bundle when exiting screen mSesState.
	 *     Lock/Unlock the input fields.
	 *     Update button icon to unlock or save/cancel.
	 */
	OnClickListener clickerEditLock = new OnClickListener() {
		public void onClick(View v) {
			Log.d(LOGTAG, "EditLock onClick() 00");

			// toggle state and update EditLock button text, based on state FIXME: this should likely go into a method
			if (FLG_SES_STATE_VIEW == mSesState) {
				mSesState = FLG_SES_STATE_EDIT;
				mEditLockButton.setText(getString(R.string.button_detail_save)); // FIXME: probably need a better string identifier name
				//loop through widgets and lock/unlock
				updateSessionDetailWidgets();			
			} else if (FLG_SES_STATE_EDIT == mSesState){
				// Prompt user if they want to save or cancel changes.
            	MPQGenDialogFragment ldlg = new MPQGenDialogFragment();
            	Bundle args = new Bundle();
            	args.putInt(MPQGenDialogFragment.ARG_DIALOG_TYPE, MPQGenDialogFragment.ARG_DIALOG_TYPE_SAVE_DISCARD);
            	args.putString(MPQGenDialogFragment.ARG_DIALOG_TITLE, getString(R.string.dialog_mpqadd_title_save_discard));
            	args.putInt(MPQGenDialogFragment.ARG_DIALOG_ID, DLG_SAVE_DISCARD);
            	ldlg.setArguments(args);
            	getTag();
            	ldlg.setTargetFragment( getActivity().getSupportFragmentManager().findFragmentByTag(getString(R.string.frag_tab2_detail))  , MPQGenDialogFragment.ARG_DIALOG_TYPE_SAVE_DISCARD);
            	ldlg.show(getActivity().getSupportFragmentManager(), getString(R.string.frag_dialog_save_discard));

// FIXME: hacking -- adding dialog/listener...				
				//if yes - save - lock, update text
				//if no - don't save-discard changes - lock, update text			
//				mSesState = FLG_SES_STATE_VIEW;
//				mEditLockButton.setText(getString(R.string.button_detail_unlock));
			}
		}
	}; // END OnClickListener clickerEditLock

	
	/**
	 * MPQGenDialog Positive Click Callback.
	 * @param dialog DialogFragment
	 * @param id int to identify callback
	 * @throws IllegalArgumentException for unknown id
	 */
	public void onMPQGenDialogPositiveClick(DialogFragment dialogFrag, int id) {
		Log.d(LOGTAG, "onMPQGenDialogPOS() 00");
		
	    // Call ContentResolver with uri and save values from dialog.
		if (id == DLG_SAVE_DISCARD) {			
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 10 id SAVE DISCARD");
			// FIXME: very warty - very little data abstraction -- a detail session object is likely called for... for now, deliver something
			// FIXME: only run this/update if data has been edited by the user (should keep a[nother] state flag) ... not too easy to implement though, as it means every control needs to be watched, i believe; alternately, get DB data, and compare widgets with DB data... (still a good deal of effort)
			
			// Collect data from controls to store
			ContentValues newValues = new ContentValues();
			SessionDetailWidgetMeta lwm;
			long ltimeSec;
			String ltext;
			int lrowCount;	// Count of how many rows were updated (should be 1).
			long lrowId;
			int lcol = 1;
			EditText let; // Local EditText view handle
			Spinner lsp;  // Local Spinner handle;
			Uri luri;
			// DateStamp
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through controls
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strDateLong(ltext);
			newValues.put(SessionsDB.PRACTICE_DATE_STAMP, ltimeSec);
			// length
			lwm = mDetailWidgets.get(lcol++);
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strSessionMSecLong(ltext);
			newValues.put(SessionsDB.PRACTICE_LENGTH_TOTAL, ltimeSec);
			// noise length
			lwm = mDetailWidgets.get(lcol++);
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			ltimeSec = SessionsDB.strSessionMSecLong(ltext);
			newValues.put(SessionsDB.PRACTICE_LENGTH_NOISE, ltimeSec);
			// audio filename
			lcol++; // ugly increment-er... must be a better way... 
// FIXME: what to do with this - if anything (maybe for now, audio file is fixed to session)
			// goal 
			lwm = mDetailWidgets.get(lcol++);
			lsp = (Spinner) lwm.mUIWidget;
			lrowId = lsp.getSelectedItemId(); 
//			lrowId++; //// Translate spinner id into row i; former is indexed on 0, latter on 1.
			newValues.put(SessionsDB.PRACTICE_GOAL_ID, lrowId);
			// notes
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through
			let = (EditText) lwm.mUIWidget;
			ltext = let.getText().toString(); 	
			newValues.put(SessionsDB.PRACTICE_NOTES, ltext);
			// instrument
			lwm = mDetailWidgets.get(lcol++);
			lsp = (Spinner) lwm.mUIWidget;
			lrowId = lsp.getSelectedItemId(); 
			newValues.put(SessionsDB.PRACTICE_INSTRUMENT_ID, lrowId);
			

			luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, mSesId);
			lrowCount = getActivity().getContentResolver().update(luri, newValues, null, null);
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 18 newValues:" + newValues.toString() );
			Log.d(LOGTAG, "Detail CallBack onMPQGenDialogPositiveClick() 19  rowCount:" + lrowCount );

			// Toggle state and update locked widgets
			mSesState = FLG_SES_STATE_VIEW;
			updateSessionDetailWidgets();
			updateSessionDetailData(); 
			mEditLockButton.setText(getString(R.string.button_detail_unlock));
		} else {
            throw new IllegalArgumentException("SessionDetailFragment onMPQGenDialogPositiveClick() Unknown DLG_ id: " + id);
		}		
		dialogFrag.dismiss();
	} // END onMPQGenDialogPositiveClick()
	
	/**
	 * MPQGenDialog Negative Click Callback.  Really nothing to do, other than dismiss the dialog.
	 * @param dialogFrag DialogFragment
	 * @param id int to identify callback
	 * @throws IllegalArgumentException for unknown id
	 */
	public void onMPQGenDialogNegativeClick(DialogFragment dialogFrag, int id) {
		Log.d(LOGTAG, "onMPQGenDialogNeg() 00 id: " + id);
		if (id == DLG_SAVE_DISCARD) {
			// Toggle state and update locked widgets
			mSesState = FLG_SES_STATE_VIEW;
			updateSessionDetailWidgets();
			updateSessionDetailData(); // FIXME: only run this if data has been edited by the user (should keep a[nother] state flag) ... not too easy to implement though, as it means every control needs to be watched, i believe
			mEditLockButton.setText(getString(R.string.button_detail_unlock));
		} else {
            throw new IllegalArgumentException("SessionDetailFragment onMPQGenDialogPositiveClick() Unknown DLG_ id: " + id);
		}
		dialogFrag.dismiss();
// FIXME: we likely want to call a common worker method, depending on the id		
	}	
	
	// ///////////////////////////////////////////////////
	// LoaderManager.LoaderCallbacks<Cursor> implementations
	/**
	 * LoaderManager callback to create new loaders.
	 */
	public Loader<Cursor> onCreateLoader(int id, Bundle args) {
		Log.d(LOGTAG, "onCreateLoader() 00 id: " + id);

		String[] lproj;
		String lselc = null;
		String[] lsela = { "" };
		Uri luri;
		
		// Setup projection		
		if (LDR_INSTRUMENTS == id) {
			Log.d(LOGTAG, "onCreateLoader() 10 id: " + id);
			lproj = new String[] { SessionsDB._ID, SessionsDB.INSTRUMENT_NAME };
			luri = SessionsDB.CONTENT_INSTRUMENTS_URI;
		} else if (LDR_GOALS == id) {
			Log.d(LOGTAG, "onCreateLoader() 12 id: " + id);
			lproj = new String[] { SessionsDB._ID, SessionsDB.GOAL_NAME };
			luri = SessionsDB.CONTENT_GOALS_URI;
		} else if (LDR_SESSIONS == id) {
			Log.d(LOGTAG, "onCreateLoader() 14 id: " + id);			
			SessionDetailWidgetMeta lwm;
			lproj = new String[mDetailWidgets.size()]; // temporary array for projection, the size of the number of columns we'll be getting
			for (int i=0; i < mDetailWidgets.size(); i++) { // FIXME; want to ensure we're in order of mDBColPos
				lwm = mDetailWidgets.get(i);
				lproj[i] = lwm.mDBCol;
				Log.d(LOGTAG, "onCreateLoader() 15a lproj[" + i + "]: " + lproj[i]);
			}
			luri = SessionsDB.CONTENT_PRACTICES_URI;
			luri = ContentUris.withAppendedId(luri, mSesId);
			
			// Update max row count, and store it in a member attribute.
			updateSessionDetailCount();
		} else {
			Log.d(LOGTAG, "onCreateLoader() 16 id: " + id);			
            throw new IllegalArgumentException("onCreateLoader() id: " + id);
		}

		Log.d(LOGTAG, "onCreateLoader() 20 id: " + id);		
		// Get cursor
		CursorLoader lcursLoader = new CursorLoader(getActivity(),
				luri, lproj, lselc, lsela, null);
		
		Log.d(LOGTAG, "onCreateLoader() 99");
		return lcursLoader;
		
		
		
	} // END onCreateLoader()

	/**
	 * LoaderManager callback to process when loading cursor data is completed.
	 */
	public void onLoadFinished(Loader<Cursor> loader, Cursor data) {
		int id = loader.getId();
		Log.d(LOGTAG, "onLoadFinished() 00");
		if (null == data) {
			Log.d(LOGTAG, "onLoadFinished() 02 - data is null - return now");
// FIXME: if sessions -- this is likely where to call create a new session
			return;
		}
		if (LDR_INSTRUMENTS == id) {
			int lid;
			Spinner lsp = (Spinner) getActivity().findViewById(R.id.detail_spinner_instrument);
 			// Be careful not to overwrite selected item in the spinner.
			lid = lsp.getSelectedItemPosition();
			Log.d(LOGTAG, "onLoadFinished() 02 Instrument data lid: " + lid );
			mAdapterInstruments.swapCursor(data);			// FIXME: be careful not to overwrite selected item in the spinner...
			lsp.setSelection(lid); 
		} else if (LDR_GOALS == id) {
			int lid;
			Spinner lsp = (Spinner) getActivity().findViewById(R.id.detail_spinner_goal);
 			// Be careful not to overwrite selected item in the spinner.
			lid = lsp.getSelectedItemPosition();
			mAdapterGoals.swapCursor(data);
			lsp.setSelection(lid); 
		} else if (LDR_SESSIONS == id) {
			Log.d(LOGTAG, "onLoadFinished() 10 Session data columns: " + Arrays.toString(data.getColumnNames()) );
			Log.d(LOGTAG, "onLoadFinished() 12 Session data pos: " + data.getPosition() + " count: " + data.getCount() + " columnCount: " + data.getColumnCount() + " mSesId:" + mSesId);
			// FIXME: very warty - very little data abstraction -- a detail session object is likely called for... for now, deliver something

			// populate controls, ideally these would be looped through consistently -- blocking is the variety in data/widget types...
			SessionDetailWidgetMeta lwm;
			long ltimeSec;
			String ltext;
			int lid;
			int lcol = 1;
			EditText let; // local EditText view handle
			TextView ltv; // local TextView handle;
			Spinner lsp;  // local Spinner handle;
			// DateStamp
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through controls
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(SessionsDB.longDateString(ltimeSec)); 		
			// length
			lwm = mDetailWidgets.get(lcol++);
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText( SessionsDB.longSessionMSecString(ltimeSec));		
			// noise length
			lwm = mDetailWidgets.get(lcol++);
			ltimeSec = data.getLong(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(SessionsDB.longSessionMSecString(ltimeSec));					
			// audio filename 
// FIXME: how to generate/deal with empty/new filename associated -- need method - maybe default to something around the datestamp
			lwm = mDetailWidgets.get(lcol++);
			ltext = data.getString(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 14 filename:" + ltext + ":");
			updateFileName(ltext);
			ltv = (TextView) lwm.mUIWidget;
			//FIXME: error-taking over screen			ltv.setText("wave-form visualization placeholder file: " + ltext);											
			ltv.setText("audio file: " + ltext + ":  sessions _ID:" + data.getLong(0) + ": min/Max-count :" + mSesMin + "/" + mSesMax + "-" + mSesCount);
			lwm = mDetailWidgets.get(lcol++);
			lid = data.getInt(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 16 Session DB lid: " + lid + " col: " + lwm.mDBCol );
			if (lid > 0) lid--; // Translate row id into spinner id; former is indexed on 1, latter on 0.
			lsp = (Spinner) lwm.mUIWidget;
			lsp.setPrompt("Goals:");
			lsp.setSelection(lid); 
			// notes
			lwm = mDetailWidgets.get(lcol++); // FIXME: horrible way to loop through
			ltext = data.getString(lwm.mDBColPos);
			let = (EditText) lwm.mUIWidget;
			let.setText(ltext);								
			// instrument
			lwm = mDetailWidgets.get(lcol++);
			lid = data.getInt(lwm.mDBColPos);
			Log.d(LOGTAG, "onLoadFinished() 18 Session DB lid: " + lid + " col: " + lwm.mDBCol );
			if (lid > 0) lid--; // translate row id into spinner id; former is indexed on 1, latter on 0
			lsp = (Spinner) lwm.mUIWidget;
			lsp.setPrompt("Instruments:");
			lsp.setSelection(lid);
		} else {
            throw new IllegalArgumentException("onFinishedLoader() getId: " + id);
		}
		Log.d(LOGTAG, "onLoadFinished() 99");
		// The list should now be shown. -- FIXME: is this "now be shown"
		// necessary???
		// if (isResumed()) {
		// setListShown(true);
		// } else {
		// setListShownNoAnimation(true);
		// }
	} // END onLoadFInished()

	/**
	 * LoaderManager callback when releasing data.
	 */
	public void onLoaderReset(Loader<Cursor> loader) {
		int id = loader.getId();
		Log.d(LOGTAG, "onLoaderReset() 00");
		if (LDR_INSTRUMENTS == id) {
			mAdapterInstruments.swapCursor(null);
		} else if (LDR_GOALS == id) {
			mAdapterGoals.swapCursor(null);
		}  else if (LDR_SESSIONS == id) {
			; // pass for now - data was copied into static widgets - no persistent cursor to maintain.  - should data be written back to disk here? does not feel appropriate...
		} else  {
            throw new IllegalArgumentException("onLoaderReset() getId: " + id);
		}
	} // END onLoaderReset()


	// //////////////////////////////
	// worker methods for record/transport control FIXME: these belong in a
	// dedicated class
	//
	
	/**
	 * Utility method to correctly update the member attribute storing the filename. 
	 * @param fn String representing the path to the file on the SD device. ex: "/mpq/file01.3gp"
	 */
	private void updateFileName(String fn) {
		String lDir = Environment.getExternalStorageDirectory().getAbsolutePath();
		if ( (null == fn) || (0 == fn.length()) ) {
			mFileName = "no audio recorded";
		} else { // FIXME: probably should use dedicated method to combine filenames... not sure what that is now
			if (fn.startsWith("/")) {
				mFileName = lDir + fn;
			} else {
				mFileName = lDir + "/" + fn;				
			}
		}
		Log.d(LOGTAG, "updateFileName() 99 fn:" + fn + ": mFileName:" + mFileName);
	} // END updateFileName(String)
	
	/**
	 * Utility method to re-query the Practice Sessions table.  This is used when browsing through records.
	 */
	private void updateSessionDetailData() { // FIXME: very warty ... little sense of clear abstraction...
		getLoaderManager().restartLoader(LDR_SESSIONS, null, this);
	}
	
	/**
	 * Utility method to loop through UI widgets, and enable or disable them 
	 * based on session state.
	 * @global mSesState is read for session sate
	 * @global mDetailWidgets is read for widgets to update
	 */
	private void updateSessionDetailWidgets() {
		Log.d(LOGTAG, "updateSessionDetailWidgets() 00 state: " + mSesState);
		
		SessionDetailWidgetMeta lwm; // Local widget meta object.
		View lev; 	// Local View handle.

		//loop through widgets and lock/unlock
		for (int i = 0; i < mDetailWidgets.size(); i++ ) {
			lwm = mDetailWidgets.get(i);
			if ( (SessionDetailWidgetMeta.WT_EDITTEXT == lwm.mWidgetType) ||
					(SessionDetailWidgetMeta.WT_SPINNER == lwm.mWidgetType) ) {
				lev = (View) lwm.mUIWidget;
				// Toggle enabled state appropriately.
				if (FLG_SES_STATE_VIEW == mSesState) {
					lev.setEnabled(false);
				} else {
					lev.setEnabled(true);
				}
			}
		} // END for widget loop
		
		if (FLG_SES_STATE_VIEW == mSesState) {
			mPrevButton.setEnabled(true);
			mNextButton.setEnabled(true);
		} else {
			mPrevButton.setEnabled(false);
			mNextButton.setEnabled(false);
		}

// FIXME: dig into this -- it's not working :(		
//		// Disable the keyboard if state is to view.
//		if (FLG_SES_STATE_VIEW == mSesState) {
//			InputMethodManager imm = (InputMethodManager) getActivity().getSystemService(Context.INPUT_METHOD_SERVICE);
//			imm.hideSoftInputFromWindow(getView().getWindowToken(), 0);
//		}
	} // END updateSessionDetailWidgets()
	
	/**
	 * Helper method to query the DB and update the min and max values and store them in member attributes.
	 */
	private void updateSessionDetailCount() {
		Cursor lcur = getActivity().getContentResolver().query(SessionsDB.CONTENT_PRACTICES_COUNT_URI, null, null, null, null);
		Log.d(LOGTAG, "updateSessionDetailCount() 10   old mSesMin: " + mSesMin + " mSesMax: " + mSesMax + " mSesCount: " + mSesCount);
		Log.d(LOGTAG, "updateSessionDetailCount() 15   mid count(): " + lcur.getCount() );
		mSesCount = lcur.getLong(0);
		mSesMin = lcur.getLong(1);
		mSesMax = lcur.getLong(2);
		Log.d(LOGTAG, "updateSessionDetailCount() 90   NEW mSesMin: " + mSesMin + " mSesMax: " + mSesMax + " mSesCount: " + mSesCount);
		lcur.close();
	}

	/**
	 * Helper method to query the DB for a session.  
	 * Returns number of rows with the session 1 or 0.
	 */
	private int getSessionDetailIdCount(long rowId) {
		int li = 0;
		Uri luri = ContentUris.withAppendedId(SessionsDB.CONTENT_PRACTICES_URI, rowId);
		Cursor lcur = getActivity().getContentResolver().query(luri, null, null, null, null);
		if (null != lcur) {
			li = lcur.getCount();
		}
		Log.d(LOGTAG, "getSessionDetailIdCount() 15  count(): " + li );
		lcur.close();		
		return li;
	}
	
	
	/**
	 * Utility method to create a new practice session, and display it.
	 * @returns long with Row ID of the new row
	 */
	private long newSessionDetailData() {
		// FIXME: should probably check that a practice session is currently already in-process...
		
		SimpleDateFormat fileDateFormat = new SimpleDateFormat("yyyyMMdd_hhmm"); // FIXME: this should likely be moved to strings.xml
		long lNowSec = java.lang.System.currentTimeMillis();
		
		// calculate filename, based on date-stamp
		SharedPreferences sharedPref = PreferenceManager.getDefaultSharedPreferences(getActivity());
		String lDirSD = sharedPref.getString(getString(R.string.pref_audiodirectory_key), "/mpq/"); // FIXME: move to strings.xml
		
		// FIXME: should add more checks: available disk space check, consumed disk space check, total size used check
		String lDir = Environment.getExternalStorageDirectory().getAbsolutePath();
		String lFileName = lDirSD + "session_";		
		lFileName = lFileName + fileDateFormat.format(new Date(lNowSec)) + ".3gp";
		
		String lFileNameFull = lDir + "/" + lFileName;
		File lf = new File(lFileNameFull);
		
		// insert new session - with filename		
        ContentValues val = new ContentValues();
        val.put(SessionsDB.PRACTICE_DATE_STAMP, lNowSec);
        val.put(SessionsDB.PRACTICE_AUDIO_FILE, lFileName);
        Uri lNewUri;
		lNewUri = getActivity().getContentResolver().insert(SessionsDB.CONTENT_PRACTICES_URI, val);
		long lNewRow = ContentUris.parseId(lNewUri);
		Log.d(LOGTAG, "newSessionDetailData() 99 lNewRow: " + lNewRow);
		if (lNewRow > 0) {
			mSesId = lNewRow;
		}
		return mSesId;
	} // END newSessionDetailData()
	
	
	private void startPlaying() {
		mPlayer = new MediaPlayer();
		try {
			mPlayer.setDataSource(mFileName);
			mPlayer.prepare();
			mPlayer.start();
		} catch (IOException e) {
			String lmsg = "startPlaying() failed. FileName:" + mFileName;
			Toast.makeText(getActivity(),lmsg, Toast.LENGTH_LONG).show();
			Log.e(LOGTAG, lmsg);
		}
	}

	private void stopPlaying() {
		mPlayer.release();
		mPlayer = null;
	}

	private void startRecording() {
		mRecorder = new MediaRecorder();
		mRecorder.setAudioSource(MediaRecorder.AudioSource.MIC);
		mRecorder.setOutputFormat(MediaRecorder.OutputFormat.THREE_GPP);
		mRecorder.setOutputFile(mFileName);
		mRecorder.setAudioEncoder(MediaRecorder.AudioEncoder.AMR_NB); // pretty much the only option to API 8 (android 2.2)
		// Not an option with AMR_NB, fixed at 8kHz -- mRecorder.setAudioSamplingRate(44100);
		mRecorder.setAudioEncodingBitRate(8000); // FIXME: test... or is this 8 vs 16 bit encoding?

		try { // curious - fewer steps inside try/catch here
			mRecorder.prepare();
		} catch (IOException e) {
			String lmsg = "mRecorder.prepare() failed. FileName:" + mFileName;
			Toast.makeText(getActivity(),lmsg, Toast.LENGTH_LONG).show();
			Log.e(LOGTAG, lmsg);
		}
		mRecorder.start();
	}

	private void stopRecording() {
		mRecorder.stop();
		mRecorder.release();
		mRecorder = null;
	}

	// ///////////////
	// needless extra step methods (workers to call other workers) -- these seem
	// very redundant for play vs. record function FIXME: this is
	// record/transport control stuff -- move this to its own class.
	//
	private void onRecord(boolean start) {
		if (start) {
			startRecording();
		} else {
			stopRecording();
		}
	}

	private void onPlay(boolean start) {
		if (start) {
			startPlaying();
		} else {
			stopPlaying();
		}
	}

}
